-- Copyright (C) 2010 Noé Falzon

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

require "luagl"

local oldprint = print
local TI = table.insert

module(..., package.seeall)


local function split(line)
	local words = {}
	for w in string.gmatch(line, "%S+") do
		table.insert(words, w)
	end
	
	words.orig = line
	
	return words
end


function load(filename)
	
	local file = io.open(filename, "r")
	local obj = {
		v = {},
		vn = {},
		vt = {},
		f = {},
		fn = {}		--face normals
	}
	
	for line in file:lines() do
		w = split(line)
		
		if w[1] == 'v' then
			TI(obj.v, {tonumber(w[2]), tonumber(w[3]), tonumber(w[4])})
		elseif w[1] == 'vn' then
			TI(obj.vn, {tonumber(w[2]), tonumber(w[3]), tonumber(w[4])})
		elseif w[1] == 'vt' then
			TI(obj.vt, {tonumber(w[2]), tonumber(w[3])})
		elseif w[1] == 'f' then
			local face = {}
			for a,b,c in w.orig:gmatch("(%d*)/(%d*)/(%d*)") do
				TI(face, {tonumber(a), tonumber(b), tonumber(c)})
			end
			TI(obj.f, face)
		end
	end
	
	-- make face normals, assume CCW vertex order.
	for i,f in ipairs(obj.f) do
		local v1 = obj.v[f[1][1]]
		local v2 = obj.v[f[2][1]]
		local v3 = obj.v[f[3][1]]
		
		u = { v3[1] - v2[1], v3[2] - v2[2], v3[3] - v2[3] }	-- 3 - 2
		v = { v1[1] - v2[1], v1[2] - v2[2], v1[3] - v2[3] }	-- 1 - 2
		
		-- u x v
		local n = {
			u[2]*v[3] - u[3]*v[2],
			u[3]*v[1] - u[1]*v[3],
			u[1]*v[2] - u[2]*v[1]
		}
		
		-- normalize
		local norm = math.sqrt(n[1]^2 + n[2]^2 + n[3]^2)
		n[1] = n[1]/norm
		n[2] = n[2]/norm
		n[3] = n[3]/norm
		
		
		obj.fn[i] = n
	end
	
	--makeDisplayList(obj)
	
	return obj
end

function print(truc)
	
	oldprint 'verts'
	for i,v in ipairs(truc.v) do
		oldprint(v[1], v[2], v[3])
	end

	oldprint 'norms'
	for i,v in ipairs(truc.vn) do
		oldprint(v[1], v[2], v[3])
	end

	oldprint 'faces'
	for i,v in ipairs(truc.f) do
		oldprint("face " .. i)
		for j,w in ipairs(v) do
			oldprint(w[1], w[2], w[3])
		end
	end
end

function makeDisplayList(obj, smooth, tex)
	
	local smooth = smooth --or true	-- by default, use model's vertices
	
	if not obj.dlist then
		
		obj.dlist = gl.GenLists(1)
		gl.NewList(obj.dlist, 'COMPILE')
		
		for i,f in ipairs(obj.f) do
			gl.Begin('POLYGON')
			
			-- in flat, use face normals
			if not smooth then
				gl.Normal(obj.fn[i])
			end
			
			for j,v in ipairs(f) do
				
				-- in smooth, use vertex normals
				if smooth then
					gl.Normal(obj.vn[v[3]])
				end
				if tex then gl.TexCoord(obj.vt[v[2]]) end
				
				gl.Vertex(obj.v[v[1]])
			end
			gl.End()
		end
		gl.EndList()
	end
end

function draw(obj, smooth, tex)
	
	for i,f in ipairs(obj.f) do
		gl.Begin('POLYGON')
		
		-- in flat, use face normals
		if not smooth then
			gl.Normal(obj.fn[i])
		end
		
		for j,v in ipairs(f) do
			
			-- in smooth, use vertex normals
			if smooth then
				gl.Normal(obj.vn[v[3]])
			end
			if tex then gl.TexCoord(obj.vt[v[2]]) end
			
			gl.Vertex(obj.v[v[1]])
		end
		gl.End()
	end
end

function drawDisplayList(obj)

	gl.CallList(obj.dlist)
end
